using System;
using System.Threading;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using static System.Diagnostics.Stopwatch;
using  GPUTools.Skinner.Scripts.Providers;
using GPUTools.Cloth.Scripts.Runtime.Data;
using GPUTools.Common.Scripts.PL.Tools;
using Leap.Unity;
using Battlehub.RTCommon;
using UnityEngine.UI;
using GPUTools.Cloth.Scripts.Geometry.Data;
using GPUTools.Cloth.Scripts.Types;
using GPUTools.Common.Scripts.Utils;


namespace Stopper {

	public class ClothCUA : MVRScript {
        string savePath = "Saves\\ClothCUA";
	    string presetFileExt = "cuaclothsettings";
        string geometryFileExt = "cuaclothgeo";

        List<GameObject> createdObjects = new List<GameObject>();

        List<ClothMeshControl> clothMeshControls = new List<ClothMeshControl>();
        List<ClothCUAMaterialOptions> materialOptions = new List<ClothCUAMaterialOptions>();

        List<MVRScriptExtensions.LayoutPanel> controlPanels = new List<MVRScriptExtensions.LayoutPanel>();

        Color buttonColor = new Color(0.839f, 0.839f, 0.839f);

        Transform mainUI;
        MVRScriptExtensions.LayoutPanel mainPanel;

        ComputeShader gpuSkinner;
        ComputeShader gpuMeshCompute;
        GameObject materialUIPrefab;
        
        JSONClass savedJson;
        CustomUnityAssetLoader loader;

        public class ClothMeshControl {
            public string name;
            public string storeId;
            public bool initialized;
            public MeshFilter meshFilter;
            public MeshRenderer meshRenderer;
            public Mesh mesh;

            public ClothSettings clothSettings;
            public CUADAZSkinV2 skin;

            public GameObject container;

            public float cuaScale;

            public MVRScriptExtensions.LayoutPanel controlPanel;
            public MVRScriptExtensions.LayoutPanel matOptsPanel;
            public Dictionary<Material,MVRScriptExtensions.LayoutPanel> matOptsPanels;
            public MVRScriptExtensions.TextBox status;

            public JSONStorableBool initOnStartup;

            public UIDynamicButton selectButton;
            public UIDynamicButton initButton;

            public JSONStorableUrl geometryDataUrl;

            public JSONStorableBool simEnabled;
            public JSONStorableBool accelerationEnabled;
            public JSONStorableBool collisionEnabled;

            public JSONStorableFloat iterations;
            
            public JSONStorableFloat uniformSimTextureStrength;

            public JSONStorableFloat stiffness;
            public JSONStorableFloat distanceScale;
            public JSONStorableFloat compressionResistance;
            public JSONStorableFloat skinJointStrength;

            public JSONStorableFloat gravityMultiplier;
            public JSONStorableFloat weight;
            public JSONStorableFloat drag;
            
            public JSONStorableFloat collisionRadius;
            public JSONStorableFloat collisionPower;
            public JSONStorableFloat friction;
            public JSONStorableFloat staticMultiplier;

            public JSONStorableFloat tesselation;
            public JSONStorableFloat tessPhong;
            
            public void Enable() {
                if(this.initialized) {
                    this.meshRenderer.enabled = false;
                    this.clothSettings.enabled = true;
                    this.skin.enabled = true;
                }
            }

            public void Disable() {
                if(this.initialized) {
                    if(this.meshRenderer != null) this.meshRenderer.enabled = true;
                    if(this.clothSettings != null) this.clothSettings.enabled = false;
                    if(this.skin != null) this.skin.enabled = false;
                }
            }
        }

        // The material opts we want are closest to the skinwrap, but we
        // just have the skin. 
        public class CUASkinWrap : DAZSkinWrap {
            protected DAZSkinV2 _skin;
            
            public DAZSkinV2 skin {
                get {
                    return _skin;
                } set {
                    _skin = value;
                    this.GPUmaterials = _skin.GPUmaterials;
                    this._numMaterials = _skin.GPUmaterials.Length;
                    this.simTextures = new Texture2D[this._numMaterials];
                    this._materialsWereInit = true;
                    this.startMesh = _skin.Mesh;
                }
            }

           
        }

        public class ClothCUAMaterialOptions : DAZSkinWrapMaterialOptions {
            public CUASkinWrap cuaWrap;
            public JSONStorableStringChooser shaderChooser;

            public void SetSkinWrap(CUASkinWrap cuaWrap) {
                this.skinWrap = cuaWrap;
                this.cuaWrap = cuaWrap;
            }

            protected override void SetMaterialHide(bool b) {
                base.SetMaterialHide(b);
                skinWrap.skin.FlushBuffers();
            }
        }

        public override void Init() {
            loader = containingAtom.reParentObject.GetComponentInChildren<CustomUnityAssetLoader>();
            if(loader != null) {
                loader.RegisterAssetLoadedCallback(AssetLoadedCallback);
                loader.RegisterAssetClearedCallback(AssetClearedCallback);

                // If we already have meshes then initialize immediately
                if(!containingAtom.reParentObject.GetComponentsInChildren<MeshFilter>().NullOrEmpty()) {
                    AssetLoadedCallback();
                }
            } else {
                StartCoroutine(Initialize());    
            }

            StartCoroutine(LoadGpuSkinner());

            foreach(var go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]) {
				if(go.name == "TabMaterialsAltWithSim"){
                    materialUIPrefab = go;
                }
			}
        }

        private void AssetLoadedCallback() {
            try {
                StartCoroutine(Initialize());
            } catch (Exception e) {
                SuperController.LogError($"{e}");
            }
        }

        private void AssetClearedCallback() {
            try {
                Reset();
                Destroy(mainPanel.gameObject);
            } catch (Exception e) {
                SuperController.LogError($"{e}");
            }
        }

        private IEnumerator Initialize() {
            yield return null;

            var meshFilters = containingAtom.reParentObject.GetComponentsInChildren<MeshFilter>();
            for(int i = 0; i < 100; i++){
                if(!meshFilters.NullOrEmpty()) {
                    break;
                } else {
                    meshFilters = containingAtom.reParentObject.GetComponentsInChildren<MeshFilter>();
                    yield return null;
                }
            }

            if(meshFilters.NullOrEmpty()) {
                SuperController.LogError($"Could not find meshes in this atom.");
            }
            
            int storeSuffix = 1;
            foreach(MeshFilter meshFilter in meshFilters) {
                ClothMeshControl control = new ClothMeshControl();
                clothMeshControls.Add(control);

                control.name = $"{meshFilter.name}";
                control.storeId = $"{meshFilter.name}{storeSuffix++}";

                control.meshFilter = meshFilter;
                control.mesh = meshFilter.mesh;
                control.meshRenderer = meshFilter.GetComponent<MeshRenderer>();

                control.initOnStartup = new JSONStorableBool($"{control.storeId}InitOnStartup", false);
                control.initOnStartup.setCallbackFunction = (val) => {
                    if(val) {
                        StartCoroutine(CreateCUACloth(control));
                    }
                };
                RegisterBool(control.initOnStartup);

                control.geometryDataUrl = new JSONStorableUrl($"{control.storeId}GeoDataUrl", null);
                RegisterUrl(control.geometryDataUrl);
                
                control.geometryDataUrl.setCallbackFunction = (path) => {
                    
                    if(control.clothSettings == null)
                        return;

                    if(!FileManagerSecure.FileExists(path))
                        return;

                    var bytes = FileManagerSecure.ReadAllBytes(path);
                    
                    control.clothSettings.GeometryData = new ClothGeometryData();
                    LoadGeometryData(control.clothSettings.GeometryData, new BinaryStreamReader(bytes));
                    control.clothSettings.GeometryData.IsProcessed = true;
                };

                control.iterations = new JSONStorableFloat($"{control.storeId}Iterations", 3f, 1f, 7f);
                RegisterFloat(control.iterations);

                control.simEnabled = new JSONStorableBool($"{control.storeId}SimEnabled", true);
                RegisterBool(control.simEnabled);

                control.accelerationEnabled = new JSONStorableBool($"{control.storeId}AccelerationEnabled", true);
                RegisterBool(control.accelerationEnabled);

                control.collisionEnabled = new JSONStorableBool($"{control.storeId}CollisionEnabled", true);
                RegisterBool(control.collisionEnabled);

                control.uniformSimTextureStrength = new JSONStorableFloat($"{control.storeId}UniformStrength", 0.5f, 0f, 1f);
                RegisterFloat(control.uniformSimTextureStrength);


                control.stiffness = new JSONStorableFloat($"{control.storeId}Stiffness", 0.5f, 0f, 1f);
                RegisterFloat(control.stiffness);

                control.distanceScale = new JSONStorableFloat($"{control.storeId}DistanceScale", 1f, 0f, 1f);
                RegisterFloat(control.distanceScale);

                control.compressionResistance = new JSONStorableFloat($"{control.storeId}CompressionResistance", 0.5f, 0f, 1f);
                RegisterFloat(control.compressionResistance);

                control.skinJointStrength = new JSONStorableFloat($"{control.storeId}SkinJointStrength", 0.5f, 0f, 1f);
                RegisterFloat(control.skinJointStrength);

                control.gravityMultiplier = new JSONStorableFloat($"{control.storeId}GravityMultiplier", 0f, 0f, 1f);
                RegisterFloat(control.gravityMultiplier);

                control.weight = new JSONStorableFloat($"{control.storeId}Weight", 1f, 0f, 2f);
                RegisterFloat(control.weight);

                control.drag = new JSONStorableFloat($"{control.storeId}Drag", 0.2f, 0f, 1f);
                RegisterFloat(control.drag);

                control.collisionRadius = new JSONStorableFloat($"{control.storeId}CollisionRadius", 0.01f, 0f, 0.1f);
                RegisterFloat(control.collisionRadius);

                control.collisionPower = new JSONStorableFloat($"{control.storeId}CollisionPower", 1f, 0f, 20f);
                RegisterFloat(control.collisionPower);

                control.friction = new JSONStorableFloat($"{control.storeId}Friction",0.5f, 0f, 1f);
                RegisterFloat(control.friction);

                control.staticMultiplier = new JSONStorableFloat($"{control.storeId}StaticMultiplier", 2f, 0f, 1f);
                RegisterFloat(control.staticMultiplier);


                control.tesselation = new JSONStorableFloat($"{control.storeId}Tesselation", 3f, 1f, 16f);
                RegisterFloat(control.tesselation);

                control.tessPhong = new JSONStorableFloat($"{control.storeId}TessPhong", 0.5f, 0f, 2f);
                RegisterFloat(control.tessPhong);
            }

            
            InitMainUI();

            
            if(savedJson != null) {
                RestoreFromJSON(savedJson, true, true);
            }
            yield break;
        }

        private void InitMainUI() {
            mainUI = this.rightUIContent.transform.parent;
            mainPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("MainPanel", mainUI);

            var headerPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("HeaderPanel", mainPanel);

            var savePreset = CreateButton("Save Preset");
            savePreset.button.onClick.AddListener(() =>
            {
                FileManagerSecure.CreateDirectory(savePath);
                var fileBrowserUI = SuperController.singleton.fileBrowserUI;
                fileBrowserUI.SetTitle("Save CUA Cloth Preset");
                fileBrowserUI.fileRemovePrefix = null;
                fileBrowserUI.hideExtension = false;
                fileBrowserUI.keepOpen = false;
                fileBrowserUI.fileFormat = geometryFileExt;
                fileBrowserUI.defaultPath = savePath;
                fileBrowserUI.showDirs = true;
                fileBrowserUI.shortCuts = null;
                fileBrowserUI.browseVarFilesAsDirectories = false;
                fileBrowserUI.SetTextEntry(true);
                fileBrowserUI.Show(SavePreset);
                fileBrowserUI.fileEntryField.text = containingAtom.name + "." + presetFileExt;
                fileBrowserUI.ActivateFileNameField();
            });
            this.AddToPanel(savePreset, headerPanel);
            
            var loadSettingsButton = CreateButton("Load Preset");
            loadSettingsButton.button.onClick.AddListener(() =>
            {
                FileManagerSecure.CreateDirectory(savePath);
                var shortcuts = FileManagerSecure.GetShortCutsForDirectory(presetFileExt);
                SuperController.singleton.GetMediaPathDialog(LoadPreset, presetFileExt, savePath, false, true, false, null, false, shortcuts);
            });
            this.AddToPanel(loadSettingsButton, headerPanel);

            var cuaControlPanel = this.AddUIElement<MVRScriptExtensions.VerticalSplitLayoutPanel>("CUAControlPanel", mainPanel);
            cuaControlPanel.ratio = 0.45f;

            var meshesLabel = this.AddTextBox($"Meshes", cuaControlPanel.leftPanel, TextAnchor.MiddleLeft, FontStyle.Bold);
            
            foreach(var control in clothMeshControls) {
                control.selectButton = CreateButton($"{control.name}");
                control.selectButton.button.onClick.AddListener(() => {
                    foreach(var cmc in clothMeshControls) {
                        cmc.controlPanel.gameObject.SetActive(false);
                        cmc.selectButton.buttonColor = buttonColor;
                    }
                    control.controlPanel.gameObject.SetActive(true);
                    control.selectButton.buttonColor = Color.green;
                });
                this.AddToPanel(control.selectButton,cuaControlPanel.leftPanel);

                control.controlPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>($"{control.storeId}ControlPanel", cuaControlPanel.rightPanel);
                controlPanels.Add(control.controlPanel);

                var header = this.AddTextBox($"{control.name}", control.controlPanel);
                header.UIText.fontStyle = FontStyle.Bold;
                header.UIText.alignment = TextAnchor.UpperLeft;

                control.status = this.AddTextBox($"", control.controlPanel);
                control.status.UIText.alignment = TextAnchor.MiddleLeft;
                control.status.gameObject.SetActive(false);

                control.initButton = CreateButton("Create Cloth");
                control.initButton.button.onClick.AddListener(() => {
                    control.initOnStartup.val = true;
                });
                this.AddToPanel(control.initButton, control.controlPanel);

                var loadPresetButton = CreateButton("Load Geometry");
                loadPresetButton.button.onClick.AddListener(() =>
                {
                    FileManagerSecure.CreateDirectory(savePath);
                    var shortcuts = FileManagerSecure.GetShortCutsForDirectory(geometryFileExt);
                    SuperController.singleton.GetMediaPathDialog((path) => {
                        control.geometryDataUrl.val = path;
                        if(path != null && path != "" && !control.initialized) {
                            control.initOnStartup.val = true;
                        }
                    }, geometryFileExt, savePath, false, true, false, null, false, shortcuts);
                });
                this.AddToPanel(loadPresetButton, control.controlPanel);

                control.controlPanel.gameObject.SetActive(false);
            }

        }


        private IEnumerator LoadGpuSkinner() {
        			// Check to see if the prefab already exists (there's a person in the scene)
			GameObject personPrefab = SuperController.singleton.GetCachedPrefab("a_per", "Person");
            if(personPrefab == null){
				// No people in the scene yet. Load one (without adding to scene) and register it as the prefab. 
				// This might have unintended consequences. 
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_per", "Person", typeof(GameObject));
				if (request == null)
				{
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
				yield return base.StartCoroutine(request);

				personPrefab = request.GetAsset<GameObject>();
				if (personPrefab != null)
				{
					SuperController.singleton.RegisterPrefab("a_per", "Person", personPrefab);
				} else {
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
			} 

            if(personPrefab != null) {
                foreach(var s in Resources.FindObjectsOfTypeAll(typeof(ComputeShader)) as ComputeShader[]) {
                    if(s.name == "DAZGPUSkin") {
                        this.gpuSkinner = Instantiate(s, this.transform);
                    }
                    if(s.name == "MeshGPU") {
                        this.gpuMeshCompute = Instantiate(s, this.transform);
                    }
                }
            }
        }

        private IEnumerator CreateCUACloth(ClothMeshControl control) {
            if(control.initialized) {
                yield break;
            }
            yield return new WaitUntil(() => this.gpuSkinner != null);

            
            control.initButton.gameObject.SetActive(false);
            control.status.gameObject.SetActive(true);

            control.container = new GameObject($"ClothCUA{control.storeId}");
            control.container.layer = 0;

            Transform parent = containingAtom.reParentObject.Find("object/rescaleObject");
            if(parent.childCount > 0) {
                parent = parent.GetChild(0);
            }
            control.container.transform.SetParent(parent, false);
            // control.container.transform.localScale = parent.localScale;
            createdObjects.Add(control.container);
            control.container.SetActive(false);

            DAZMeshData meshData = ScriptableObject.CreateInstance<DAZMeshData>();
            meshData.baseVertices = (Vector3[])control.mesh.vertices.Clone();
            meshData.UVVertices = meshData.baseVertices;
            meshData.OrigUV = (Vector2[])control.mesh.uv.Clone();

            List<DAZVertexMap> baseVerticesToUV = new List<DAZVertexMap>();
            for(int i = 0; i < meshData.baseVertices.Length; i++) {
                DAZVertexMap vm = new DAZVertexMap();
                vm.fromvert = i;
                vm.tovert = i;
                baseVerticesToUV.Add(vm);
            }
            meshData.baseVerticesToUVVertices = baseVerticesToUV.ToArray();

            List<MeshPoly> meshPolies = new List<MeshPoly>();
            for(int submesh = 0; submesh < control.mesh.subMeshCount; submesh++){
                var triangles = control.mesh.GetTriangles(submesh);
                for(int i = 0; i < triangles.Length/3; i++) {
                    MeshPoly meshPoly = new MeshPoly();
                    meshPoly.materialNum = submesh;
                    meshPoly.vertices = new int[]{triangles[3*i+2],triangles[3*i+1],triangles[3*i]};
                    meshPolies.Add(meshPoly);
                }
            }

            meshData.basePolyList = meshPolies.ToArray();
            meshData.UVPolyList = meshData.basePolyList;

            CUADAZMesh dazMesh = control.container.AddComponent<CUADAZMesh>();
            dazMesh.meshData = meshData;
            dazMesh.materials = new Material[control.meshRenderer.materials.Length];
            for(int i = 0; i < dazMesh.materials.Length; i++) {
                dazMesh.materials[i] = new Material(control.meshRenderer.materials[i]);
            }
          
            dazMesh.UpdateInternals();
            dazMesh.Init();

            control.meshRenderer.enabled = false;

            dazMesh.drawUVMappedMesh = false;
            
            CUADAZSkinV2 skin = control.container.AddComponent<CUADAZSkinV2>();
            control.skin = skin;
            skin.dazMesh = dazMesh;
            skin.skin = false;
            skin.draw = false;
            skin.drawInPostProcess = true;
            skin.useGeneralWeights = true;
            skin.useSmoothing = false;
            skin.allowPostSkinMorph = false;
            skin.GPUSkinner = Instantiate(gpuSkinner);
            skin.GPUMeshCompute = Instantiate(gpuMeshCompute);
            skin.useEarlyFinish = true;
            skin.skinMethod = DAZSkinV2.SkinMethod.GPU;
            skin.CopyMaterials();
            skin.InitSimpleBones(containingAtom.reParentObject);

             for(int i = 0; i < control.skin.GPUmaterials.Length; i++) {
                control.skin.GPUmaterials[i] = new Material(control.meshRenderer.materials[i]);
                // control.skin.GPUmaterials[i].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
                control.skin.GPUmaterials[i].shader = Shader.Find("Custom/Subsurface/GlossNMTessMappedFixedComputeBuff");
                control.skin.GPUmaterials[i].SetTexture("_TessTex", Texture2D.whiteTexture);
                control.skin.FlushBuffers();
            }

            
            var pos = control.container.transform.position;
            var rot = control.container.transform.rotation;
            control.container.transform.position = Vector3.zero;
            control.container.transform.rotation = Quaternion.identity;
            control.container.SetActive(true);

            
            control.container.transform.position = pos;
            control.container.transform.rotation = rot;

            foreach(var bone in skin.root.dazBones) {
                bone.disableMorph = true;
            }


            var clothSettings = control.container.AddComponent<ClothSettings>();
            control.clothSettings = clothSettings;
            clothSettings.enabled = false;
            clothSettings.EditorType = ClothEditorType.Texture;
            clothSettings.EditorStrengthType = ClothEditorType.Texture;
            
            clothSettings.MeshProvider.Type = ScalpMeshType.PreCalc;
            clothSettings.MeshProvider.PreCalcProvider = skin;

            control.cuaScale = control.container.transform.parent.localScale.x;
            clothSettings.WorldScale = control.container.transform.parent.localScale.x;

            if(control.geometryDataUrl.val != null && control.geometryDataUrl.val != "") {
                control.geometryDataUrl.setCallbackFunction(control.geometryDataUrl.val);
            } else {
                clothSettings.ProcessGeometryMainThread();
                var geometryThread = new Thread(new ThreadStart(clothSettings.ProcessGeometryThreaded));
                geometryThread.Start();
                while(geometryThread.IsAlive) {
                    control.status.text = clothSettings.GeometryData.status;
                    yield return null;
                }
            }

            control.status.gameObject.SetActive(false);
            clothSettings.CollisionEnabled = true;
            clothSettings.enabled = true;
            
            control.initialized = true;

            SyncControlUI(control);

            yield return null;
            clothSettings.Reset();
            yield break;
        }

        private void SyncControlUI(ClothMeshControl control) {
            var clothSettings = control.clothSettings;


            var saveGeometryButton = CreateButton("Save Geometry");
            saveGeometryButton.button.onClick.AddListener(() =>
            {
                FileManagerSecure.CreateDirectory(savePath);
                var fileBrowserUI = SuperController.singleton.fileBrowserUI;
                fileBrowserUI.SetTitle("Save CUA Cloth Geometry");
                fileBrowserUI.fileRemovePrefix = null;
                fileBrowserUI.hideExtension = false;
                fileBrowserUI.keepOpen = false;
                fileBrowserUI.fileFormat = geometryFileExt;
                fileBrowserUI.defaultPath = savePath;
                fileBrowserUI.showDirs = true;
                fileBrowserUI.shortCuts = null;
                fileBrowserUI.browseVarFilesAsDirectories = false;
                fileBrowserUI.SetTextEntry(true);
                fileBrowserUI.Show((path) => {
                    BinaryStreamWriter binWriter = new BinaryStreamWriter();
                    StoreGeometryData(control.clothSettings.GeometryData, binWriter);
			
			        FileManagerSecure.WriteAllBytes(path, binWriter.GetBytes());
                    
                    control.geometryDataUrl.valNoCallback = path;
                });
                fileBrowserUI.fileEntryField.text = containingAtom.name + "_" + control.meshRenderer.name + "." + geometryFileExt;
                fileBrowserUI.ActivateFileNameField();
            });
            this.AddToPanel(saveGeometryButton, control.controlPanel);

            var resetButton = CreateButton("Reset Sim");
            resetButton.button.onClick.AddListener(() => clothSettings.Reset());
            this.AddToPanel(resetButton, control.controlPanel);

            var simEnabledToggle = CreateToggle(control.simEnabled);
            simEnabledToggle.label = "Sim Enabled";
            this.AddToPanel(simEnabledToggle, control.controlPanel);
            control.simEnabled.setCallbackFunction = val => {
                if(val) {
                    control.Enable();
                } else {
                    control.Disable();
                }
            };


            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //  Materials
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            
            var materialsLavel = this.AddTextBox($"Materials", control.controlPanel, TextAnchor.MiddleLeft, FontStyle.Bold);
            control.matOptsPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>($"{control.storeId}MaterialOptionPanel", control.controlPanel);
            for(int i = 0; i < control.skin.GPUmaterials.Length; i++) {
                Material mat = control.skin.GPUmaterials[i];
                
                var matUIPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>($"{control.storeId}{mat.name}MatOpts", mainUI);

                // Material Options UI

                var backButton = CreateButton("Back", false);
                this.AddToPanel(backButton, matUIPanel);
                var backButtonRect = backButton.GetComponent<RectTransform>();
                backButtonRect.anchorMax = new Vector2(0.15f,1f);
                backButton.button.onClick.AddListener(() => {
                    matUIPanel.gameObject.SetActive(false);
                    mainPanel.gameObject.SetActive(true);
                });

                var matOptsUI = Instantiate(materialUIPrefab, matUIPanel.transform);
                matUIPanel.gameObject.SetActive(false);

                control.skin.gameObject.SetActive(false);
                ClothCUAMaterialOptions matOpts = control.skin.gameObject.AddComponent<ClothCUAMaterialOptions>();
                materialOptions.Add(matOpts);
                matOpts.overrideId = $"{control.storeId}{mat.name}";
                matOpts.UITransform = matOptsUI.transform;
                
                CUASkinWrap skinWrap = control.skin.gameObject.AddComponent<CUASkinWrap>();
                matOpts.enabled = true;
                skinWrap.skin = control.skin;
                matOpts.SetSkinWrap(skinWrap);
                matOpts.paramMaterialSlots = new int[] {i};
                matOpts.textureGroup1 = new MaterialOptionTextureGroup();
                matOpts.customTextureFolder = "Custom/Atoms";
                matOpts.materialForDefaults = mat;
                
                
                control.skin.gameObject.SetActive(true);


                // Control UI            
                var matHeader = this.AddTextBox($"{mat.name}", control.matOptsPanel, TextAnchor.MiddleLeft);

            
                var tesselation = new JSONStorableFloat($"{control.storeId}{mat.name}Tesselation",3,0,8);
                var tessPhong = new JSONStorableFloat($"{control.storeId}{mat.name}TessPhong",0.5f,0f,1f);
                var tesselationSlider = CreateSlider(tesselation, false);
                var tessPhongSlider = CreateSlider(tessPhong, false);


                matOpts.shaderChooser = new JSONStorableStringChooser(
                    "Shader",
                    new List<string>() { "Standard", "Transparent" },
                    "Standard",
                    "Shader",
                    (string type) => {
                        if (type == "Standard") {
                            mat.shader = Shader.Find("Custom/Subsurface/GlossNMTessMappedFixedComputeBuff");
                            tesselationSlider.gameObject.SetActive(true);
                            tessPhongSlider.gameObject.SetActive(true);
                        } else {
                            mat.shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
                            tesselationSlider.gameObject.SetActive(false);
                            tessPhongSlider.gameObject.SetActive(false);
                        }
                        control.skin.FlushBuffers();
                        matOpts.SetStartingValues();

                        matOpts.InitUI();
                    });
                matOpts.RegisterStringChooser(matOpts.shaderChooser);

                var matShaderPopup = CreatePopup(matOpts.shaderChooser, false);
                this.AddToPanel(matShaderPopup, control.matOptsPanel);
                var popupPanel = matShaderPopup.popup.popupPanel;

                var showMatOptsButton = CreateButton($"Material Options");
                showMatOptsButton.button.onClick.AddListener(() => {
                    matUIPanel.gameObject.SetActive(true);
                    mainPanel.gameObject.SetActive(false);
                });
                this.AddToPanel(showMatOptsButton, control.matOptsPanel);

                tesselationSlider.label = "Tesselation";
                this.AddToPanel(tesselationSlider, control.matOptsPanel);
                SetCallbackFunction(tesselation, val => {
                    if (mat.HasProperty("_Tess")) {
                        mat.SetFloat("_Tess", val);
                    }
                });

                tessPhongSlider.label = "Tess Phong";
                this.AddToPanel(tessPhongSlider, control.matOptsPanel);
                SetCallbackFunction(tessPhong, val => {
                    if (mat.HasProperty("_TessPhong")) {
                        mat.SetFloat("_TessPhong", val);
                    }
                });

                if(savedJson != null && savedJson["materials"] != null) {
                    var matOptJson = savedJson["materials"].AsObject[matOpts.storeId];
                    
                    if(matOptJson != null) {
                        // Set the shader first
                        if(matOptJson["Shader"] != null) {
                            matOpts.shaderChooser.val = matOptJson["Shader"];
                        }

                        // Restore other settings
                        matOpts.RestoreFromJSON(matOptJson.AsObject);
                    }
                }
                
            }
            

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //  Sim Settings
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            var simSettingsLabel = this.AddTextBox($"Sim Settings", control.controlPanel, TextAnchor.MiddleLeft, FontStyle.Bold);
            
            var accelerationToggle = CreateToggle(control.accelerationEnabled);
            accelerationToggle.label = "Acceleration Enabled";
            this.AddToPanel(accelerationToggle, control.controlPanel);
            control.accelerationEnabled.setCallbackFunction = val => clothSettings.IntegrateEnabled = val;

            var collisionToggle = CreateToggle(control.collisionEnabled);
            collisionToggle.label = "Collision Enabled";
            this.AddToPanel(collisionToggle, control.controlPanel);
            control.collisionEnabled.setCallbackFunction = val => clothSettings.CollisionEnabled = val;

            
            var iterationsSlider = AddSliderForMeshControl(control.iterations, "Solver Iterations", control);
            iterationsSlider.slider.wholeNumbers = true;

            var stiffnessSlider = AddSliderForMeshControl(control.stiffness, "Stiffness", control);
            var distanceScale = AddSliderForMeshControl(control.distanceScale, "Distance Scale", control);
            var compressionResistance = AddSliderForMeshControl(control.compressionResistance, "Compression Resistance", control);
            var uniformSimTextureStrength = AddSliderForMeshControl(control.uniformSimTextureStrength, "Sim Texture Strength", control);
            var skinJointStrength = AddSliderForMeshControl(control.skinJointStrength, "Skin Joint Strength", control);
            var gravityMultiplier = AddSliderForMeshControl(control.gravityMultiplier, "Gravity Multiplier", control);
            var weight = AddSliderForMeshControl(control.weight, "Weight", control);
            var drag = AddSliderForMeshControl(control.drag, "Drag", control);
            var collisionRadius = AddSliderForMeshControl(control.collisionRadius, "Collision Radius", control);
            collisionRadius.valueFormat= "F3";
            var collisionPower = AddSliderForMeshControl(control.collisionPower, "Collision Power", control);
            var friction = AddSliderForMeshControl(control.friction, "Friction", control);
            var staticMultiplier = AddSliderForMeshControl(control.staticMultiplier, "Static Multiplier", control);

            SetCallbackFunction(control.uniformSimTextureStrength, val => {
                int num = 4;
                int num2 = 4;
                Color color = new Color(val, 0f, 0f);
                Texture2D texture2D = new Texture2D(num, num2);
                for (int i = 0; i < num; i++)
                {
                    for (int j = 0; j < num2; j++)
                    {
                        texture2D.SetPixel(i, j, color);
                    }
                }
                texture2D.Apply();

                clothSettings.EditorTexture = texture2D;

                if (clothSettings.GeometryData != null)
				{
					clothSettings.GeometryData.ResetParticlesBlend();
				}
				if (clothSettings.builder != null)
				{
					if (clothSettings.builder.physicsBlend != null)
					{
						clothSettings.builder.physicsBlend.UpdateSettings();
					}
					clothSettings.Reset();
				}                   
            });       

            SetCallbackFunction(control.stiffness, val => clothSettings.Stiffness = val);           

            SetCallbackFunction(control.distanceScale, val => clothSettings.DistanceScale = val);

            SetCallbackFunction(control.compressionResistance, val => clothSettings.CompressionResistance = val);

            SetCallbackFunction(control.skinJointStrength, val => clothSettings.JointStrength = val);

            SetCallbackFunction(control.gravityMultiplier, val => clothSettings.GravityMultiplier = val);

            SetCallbackFunction(control.weight, val => clothSettings.Weight = val);

            SetCallbackFunction(control.drag, val => clothSettings.Drag = val);

            SetCallbackFunction(control.collisionRadius, val => {clothSettings.ParticleRadius = val; clothSettings.Reset();});

            SetCallbackFunction(control.collisionPower, val => clothSettings.CollisionPower = val);

            SetCallbackFunction(control.friction, val => clothSettings.Friction = val);

            SetCallbackFunction(control.staticMultiplier, val => clothSettings.StaticMultiplier = val);
        }

        private void ShowMainPanel() {

        }

        private UIDynamicSlider AddSliderForMeshControl(JSONStorableFloat storableFloat, string label, ClothMeshControl control) {
            var slider = CreateSlider(storableFloat);
            slider.label = label;
            this.AddToPanel(slider, control.controlPanel);
            return slider;
        }

        private void SetCallbackFunction(JSONStorableFloat storableFloat, JSONStorableFloat.SetFloatCallback callback) {
            storableFloat.setCallbackFunction = callback;
            storableFloat.setCallbackFunction(storableFloat.val);
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) {
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

			JSONClass materials = new JSONClass();
			foreach (var matOpts in materialOptions) {
                materials[matOpts.storeId] = matOpts.GetJSON();
			}
			jc["materials"] = materials;
			return jc;
		}

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            savedJson = jc;
		}

        private void SavePreset(string path) {
            SuperController.singleton.fileBrowserUI.fileFormat = null;
            if (string.IsNullOrEmpty(path)) {
                return;
            }

            if (!path.ToLower().EndsWith($".{presetFileExt}")) {
                path += $".{presetFileExt}";
            }

            SaveJSON(GetJSON(true, true, true), path);
        }

        private void LoadPreset(string path) {
            if (string.IsNullOrEmpty(path))
                return;

            var presetJson = (JSONClass)LoadJSON(path);

            if(presetJson != null && presetJson.Count > 0) {
                RestoreFromJSON(presetJson, true, true, null, false);
            }
        }

        public void OnDisable(){	
            foreach(var control in clothMeshControls) {
                control.Disable();
            }
        }

        public void OnEnable(){	
            foreach(var control in clothMeshControls) {
                control.Enable();
            }
        }

        private void Reset() {
            foreach(var paramName in new List<string>(this.GetAllParamAndActionNames())) {
                var param = this.GetParam(paramName);
                if(param is JSONStorableBool) {
                    DeregisterBool(param as JSONStorableBool);
                } else if(param is JSONStorableFloat) {
                    DeregisterFloat(param as JSONStorableFloat);
                } else if(param is JSONStorableString) {
                    DeregisterString(param as JSONStorableString);
                } else if(param is JSONStorableUrl) {
                    DeregisterUrl(param as JSONStorableUrl);
                } else if(param is JSONStorableStringChooser) {
                    DeregisterStringChooser(param as JSONStorableStringChooser);
                }
            }
            clothMeshControls.Clear();
            materialOptions.Clear();

            foreach(var go in createdObjects) {
                Destroy(go);
            }
        }

        public void OnDestroy(){
            Reset();
            if(loader != null) {
                loader.DeregisterAssetLoadedCallback(AssetLoadedCallback);
                loader.DeregisterAssetClearedCallback(AssetClearedCallback);
            }
        }

        // Taken from VAM's ClothGeometryData implementation
        public bool LoadGeometryData(ClothGeometryData data, BinaryStreamReader reader) {
            try {
                string a = reader.ReadString();
                if (a != "ClothGeometryData") {
                    return false;
                }
                string text = reader.ReadString();
                if (text != "1.0") {
                    return false;
                }
                int num = reader.ReadInt32();
                data.AllTringles = new int[num];
                for (int i = 0; i < num; i++) {
                    data.AllTringles[i] = reader.ReadInt32();
                }
                int num2 = reader.ReadInt32();
                data.Particles = new Vector3[num2];
                for (int j = 0; j < num2; j++) {
                    Vector3 vector;
                    vector.x = reader.ReadSingle();
                    vector.y = reader.ReadSingle();
                    vector.z = reader.ReadSingle();
                    data.Particles[j] = vector;
                }
                int num3 = reader.ReadInt32();
                data.MeshToPhysicsVerticesMap = new int[num3];
                for (int k = 0; k < num3; k++) {
                    data.MeshToPhysicsVerticesMap[k] = reader.ReadInt32();
                }
                int num4 = reader.ReadInt32();
                data.PhysicsToMeshVerticesMap = new int[num4];
                for (int l = 0; l < num4; l++) {
                    data.PhysicsToMeshVerticesMap[l] = reader.ReadInt32();
                }
                int num5 = reader.ReadInt32();
                data.JointGroups = new List<Int2ListContainer>();
                for (int m = 0; m < num5; m++) {
                    Int2ListContainer int2ListContainer = new Int2ListContainer();
                    data.JointGroups.Add(int2ListContainer);
                    int num6 = reader.ReadInt32();
                    int2ListContainer.List = new List<Int2>();
                    for (int n = 0; n < num6; n++) {
                        Int2 item = default(Int2);
                        item.X = reader.ReadInt32();
                        item.Y = reader.ReadInt32();
                        int2ListContainer.List.Add(item);
                    }
                }
                int num7 = reader.ReadInt32();
                data.StiffnessJointGroups = new List<Int2ListContainer>();
                for (int num8 = 0; num8 < num7; num8++) {
                    Int2ListContainer int2ListContainer2 = new Int2ListContainer();
                    data.StiffnessJointGroups.Add(int2ListContainer2);
                    int num9 = reader.ReadInt32();
                    int2ListContainer2.List = new List<Int2>();
                    for (int num10 = 0; num10 < num9; num10++) {
                        Int2 item2 = default(Int2);
                        item2.X = reader.ReadInt32();
                        item2.Y = reader.ReadInt32();
                        int2ListContainer2.List.Add(item2);
                    }
                }
                int num11 = reader.ReadInt32();
                data.NearbyJointGroups = new List<Int2ListContainer>();
                for (int num12 = 0; num12 < num11; num12++) {
                    Int2ListContainer int2ListContainer3 = new Int2ListContainer();
                    data.NearbyJointGroups.Add(int2ListContainer3);
                    int num13 = reader.ReadInt32();
                    int2ListContainer3.List = new List<Int2>();
                    for (int num14 = 0; num14 < num13; num14++) {
                        Int2 item3 = default(Int2);
                        item3.X = reader.ReadInt32();
                        item3.Y = reader.ReadInt32();
                        int2ListContainer3.List.Add(item3);
                    }
                }
                int num15 = reader.ReadInt32();
                data.ParticleToNeibor = new int[num15];
                for (int num16 = 0; num16 < num15; num16++) {
                    data.ParticleToNeibor[num16] = reader.ReadInt32();
                }
                int num17 = reader.ReadInt32();
                data.ParticleToNeiborCounts = new int[num17];
                for (int num18 = 0; num18 < num17; num18++) {
                    data.ParticleToNeiborCounts[num18] = reader.ReadInt32();
                }
                int num19 = reader.ReadInt32();
                data.ParticlesBlend = new float[num19];
                for (int num20 = 0; num20 < num19; num20++) {
                    data.ParticlesBlend[num20] = reader.ReadSingle();
                }
                int num21 = reader.ReadInt32();
                data.ParticlesStrength = new float[num21];
                for (int num22 = 0; num22 < num21; num22++) {
                    data.ParticlesStrength[num22] = reader.ReadSingle();
                }
            } catch (Exception arg) {
                SuperController.LogError($"Unable to load cloth geometry {arg}");
                return false;
            }
            return true;
        }

        public bool StoreGeometryData(ClothGeometryData data, BinaryStreamWriter writer) {
            try {
                writer.Write("ClothGeometryData");
                writer.Write("1.0");
                writer.Write(data.AllTringles.Length);
                for (int i = 0; i < data.AllTringles.Length; i++) {
                    writer.Write(data.AllTringles[i]);
                }
                writer.Write(data.Particles.Length);
                for (int j = 0; j < data.Particles.Length; j++) {
                    writer.Write(data.Particles[j].x);
                    writer.Write(data.Particles[j].y);
                    writer.Write(data.Particles[j].z);
                }
                writer.Write(data.MeshToPhysicsVerticesMap.Length);
                for (int k = 0; k < data.MeshToPhysicsVerticesMap.Length; k++) {
                    writer.Write(data.MeshToPhysicsVerticesMap[k]);
                }
                writer.Write(data.PhysicsToMeshVerticesMap.Length);
                for (int l = 0; l < data.PhysicsToMeshVerticesMap.Length; l++) {
                    writer.Write(data.PhysicsToMeshVerticesMap[l]);
                }
                writer.Write(data.JointGroups.Count);
                for (int m = 0; m < data.JointGroups.Count; m++) {
                    Int2ListContainer int2ListContainer = data.JointGroups[m];
                    writer.Write(int2ListContainer.List.Count);
                    for (int n = 0; n < int2ListContainer.List.Count; n++) {
                        writer.Write(int2ListContainer.List[n].X);
                        writer.Write(int2ListContainer.List[n].Y);
                    }
                }
                writer.Write(data.StiffnessJointGroups.Count);
                for (int num = 0; num < data.StiffnessJointGroups.Count; num++) {
                    Int2ListContainer int2ListContainer2 = data.StiffnessJointGroups[num];
                    writer.Write(int2ListContainer2.List.Count);
                    for (int num2 = 0; num2 < int2ListContainer2.List.Count; num2++) {
                        writer.Write(int2ListContainer2.List[num2].X);
                        writer.Write(int2ListContainer2.List[num2].Y);
                    }
                }
                writer.Write(data.NearbyJointGroups.Count);
                for (int num3 = 0; num3 < data.NearbyJointGroups.Count; num3++) {
                    Int2ListContainer int2ListContainer3 = data.NearbyJointGroups[num3];
                    writer.Write(int2ListContainer3.List.Count);
                    for (int num4 = 0; num4 < int2ListContainer3.List.Count; num4++) {
                        writer.Write(int2ListContainer3.List[num4].X);
                        writer.Write(int2ListContainer3.List[num4].Y);
                    }
                }
                writer.Write(data.ParticleToNeibor.Length);
                for (int num5 = 0; num5 < data.ParticleToNeibor.Length; num5++) {
                    writer.Write(data.ParticleToNeibor[num5]);
                }
                writer.Write(data.ParticleToNeiborCounts.Length);
                for (int num6 = 0; num6 < data.ParticleToNeiborCounts.Length; num6++) {
                    writer.Write(data.ParticleToNeiborCounts[num6]);
                }
                writer.Write(data.ParticlesBlend.Length);
                for (int num7 = 0; num7 < data.ParticlesBlend.Length; num7++) {
                    writer.Write(data.ParticlesBlend[num7]);
                }
                writer.Write(data.ParticlesStrength.Length);
                for (int num8 = 0; num8 < data.ParticlesStrength.Length; num8++) {
                    writer.Write(data.ParticlesStrength[num8]);
                }
            } catch (Exception arg) {
                SuperController.LogError($"Unable to save cloth geometry {arg}");
                return false;
            }
            return true;
        }

        public class CUADAZMesh : DAZMesh {
            
            public void UpdateInternals() {
                this._numBaseVertices = this.meshData.baseVertices.Length;
                this._numUVVertices = this.meshData.UVVertices.Length;
                this._numBasePolygons = this.meshData.basePolyList.Length;
                this._numMaterials = this.materials.Length;

                this.materialsEnabled = new bool[this.materials.Length];
                this.materialsShadowCastEnabled = new bool[this.materials.Length];
                this._materialNames = new string[this.materials.Length];
                for(int i = 0; i < this.materials.Length; i++) {
                    this.materialsEnabled[i] = true;
                    this.materialsShadowCastEnabled[i] = true;
                    this._materialNames[i] = this.materials[i].name;
                }
            }
        }

        public class CUADAZSkinV2 : DAZSkinV2 {

            public void InitSimpleBones(Transform reParentObject) {
                GameObject rootBoneGo = new GameObject("root");
                rootBoneGo.transform.SetParent(this.transform,true);
                rootBoneGo.transform.localPosition = Vector3.zero;
                rootBoneGo.transform.localRotation = Quaternion.identity;
                
                DAZBone rootBone = rootBoneGo.AddComponent<DAZBone>();
                rootBone.disableMorph = false;
                rootBone.enabled = true;
                rootBone.isRoot = true;

                DAZBones root = this.gameObject.AddComponent<DAZBones>();
                root.dazBones = new DAZBone[]{rootBone};
                // root.useScale = true;
                
                DAZSkinV2Node node = new DAZSkinV2Node();
                node.name = rootBone.name;
                node.url = node.name;
                node.id = node.name;

                // We'll use general weights
                node.weights = new DAZSkinV2VertexWeights[0];
                node.fullyWeightedVertices = new int[0];
                node.generalWeights = new DAZSkinV2GeneralVertexWeights[this.dazMesh.numBaseVertices];
                for(int i = 0; i < node.generalWeights.Length; i++) {
                    DAZSkinV2GeneralVertexWeights generalWeight = new DAZSkinV2GeneralVertexWeights();
                    generalWeight.vertex = i;
                    generalWeight.weight = 1;
                    node.generalWeights[i] = generalWeight;
                }
                
                DAZSkinV2BulgeFactors bulgeFactors = new DAZSkinV2BulgeFactors();
                bulgeFactors.name = "root";
                node.bulgeFactors = bulgeFactors;

                nodes = new DAZSkinV2Node[]{node};
                rootBone.dazBones = root;
                this.dazBones = new DAZBone[] {rootBone};
                this.root = root;
                this._numBones = 1;
            }

            public override Mesh MeshForImport {
                get {
                    return Mesh;
                }
            }

            public override GpuBuffer<Vector3> PreCalculatedVerticesBuffer {
                get {
                    var buffer = base.PreCalculatedVerticesBuffer;
                    if(buffer.Data == null) {
                        buffer = new GpuBuffer<Vector3>(this.rawVertsBuffer);
                        this.PreCalculatedVerticesBuffer = buffer;
                        buffer.Data = new Vector3[rawVertsBuffer.count];
                    }
                    return buffer;
                }
                protected set {
                    base.PreCalculatedVerticesBuffer = value;
                }
            }


            // Overwriting this to use the rawvertsbuffer
            public override void Dispatch() {
                if (this.needsDispatch && this.rawVertsBuffer != null) {
                    this.SkinMeshGPU();
                    this.needsDispatch = false;
                    this.GPUSkinner.SetBuffer(this._copyKernel, "inVerts", this.rawVertsBuffer);
                    this.GPUSkinner.SetBuffer(this._copyKernel, "outVerts", this.preCalcVertsBuffer);
                    this.GPUSkinner.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                    if (this.provideToWorldMatrices) {
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "originalPositions", this._originalVerticesBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "originalNormals", this._originalNormalsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "originalTangents", this._originalTangentsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "currentPositions", this.preCalcVertsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "currentNormals", this._normalsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "currentTangents", this._tangentsBuffer);
                        this.GPUSkinner.SetBuffer(this._calcChangeMatricesKernel, "vertexChangeMatrices", this._matricesBuffer);
                        this.GPUSkinner.Dispatch(this._calcChangeMatricesKernel, this.numVertThreadGroups, 1, 1);
                    }
                }
            }


            public override void PostProcessDispatch(ComputeBuffer finalVerts) {
                if (this.drawInPostProcess && this.preCalcVertsBuffer != null) {
                    this.GPUSkinner.SetBuffer(this._copyKernel, "inVerts", finalVerts);
                    this.GPUSkinner.SetBuffer(this._copyKernel, "outVerts", this.rawVertsBuffer);
                    this.GPUSkinner.Dispatch(this._copyKernel, this.numVertThreadGroups, 1, 1);
                   
                    this.DrawMeshGPU();
                }
            }
        }
    }
}